import base58
from LTO import crypto
from LTO.transaction import Transaction
import struct

class RevokeAssociation(Transaction):
    TYPE = 17
    DEFAULT_LEASE_FEE = 100000000
    DEFAULT_VERSION = 3

    def __init__(self, recipient, association_type, anchor = ''):
        super().__init__()
        self.recipient = recipient
        self.anchor = anchor
        self.association_type = association_type

        self.tx_fee = self.DEFAULT_LEASE_FEE
        self.version = self.DEFAULT_VERSION

    def __to_binary_V1(self):
        if self.anchor:
            return (self.TYPE.to_bytes(1, 'big') +
                    b'\1' +
                    crypto.str2bytes(crypto.get_network(self.sender)) +
                    base58.b58decode(self.sender_public_key) +
                    base58.b58decode(self.recipient) +
                    struct.pack(">i", self.association_type) +
                    b'\1' +
                    struct.pack(">H", len(crypto.str2bytes(self.anchor))) +
                    crypto.str2bytes(self.anchor) +
                    struct.pack(">Q", self.timestamp) +
                    struct.pack(">Q", self.tx_fee))
        else:
            return (self.TYPE.to_bytes(1, 'big') +
                    b'\1' +
                    crypto.str2bytes(crypto.get_network(self.sender)) +
                    base58.b58decode(self.sender_public_key) +
                    base58.b58decode(self.recipient) +
                    struct.pack(">i", self.association_type) +
                    b'\0' +
                    struct.pack(">Q", self.timestamp) +
                    struct.pack(">Q", self.tx_fee))


    def __to_binary_V3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                base58.b58decode(self.recipient) +
                struct.pack(">i", self.association_type) +
                struct.pack(">H", len(crypto.str2bytes(self.anchor))) +
                crypto.str2bytes(self.anchor)
                )

    def to_binary(self):
        if self.version == 1:
            return self.__to_binary_V1()
        elif self.version == 3:
            return self.__to_binary_V3()
        else:
            raise Exception('Incorrect Version')

    def to_json(self):
        return ({
            "type": self.TYPE,
            "version": self.version,
            "sender": self.sender_public_key,
            "senderKeyType": self.sender_key_type,
            "senderPublicKey": self.sender_public_key,
            "recipient": self.recipient,
            "associationType": self.association_type,
            "hash": base58.b58encode(crypto.str2bytes(self.anchor)),
            "timestamp": self.timestamp,
            "fee": self.tx_fee,
            "proofs": self.proofs
        } | self._sponsor_json())

    @staticmethod
    def from_data(data):
        tx = RevokeAssociation(recipient='', association_type='')
        tx.type = data['type']
        tx.version = data['version']
        tx.id = data['id'] if 'id' in data else ''
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.recipient = data['party'] if 'party' in data else data['recipient']
        tx.association_type = data['associationType']
        tx.hash = data['hash'] if 'hash' in data else ''
        tx.timestamp = data['timestamp']
        tx.fee = data['fee']
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''
        return tx


from LTO.account_factory import AccountFactory
from LTO import crypto
import struct
from nacl.signing import SigningKey, VerifyKey
import base58
from LTO.account import Account
from LTO.Accounts.ED25519.account_ed25519 import AccountED25519 as Account
# from LTO.Accounts.ED25519 import SigningKeyED25519 as SigningKey


class AccountFactoryED25519(AccountFactory):

    def __init__(self, chain_id):
        super().__init__(chain_id)
        self.key_type = 'ed25519'

    def create_sign_keys(self, seed, nonce=0):
        seed_hash = crypto.hash_chain(struct.pack(">L", nonce) + crypto.str2bytes(seed))
        account_seed_hash = crypto.sha256(seed_hash)
        private_key = SigningKey(account_seed_hash)
        public_key = private_key.verify_key
        return private_key, public_key, self.key_type

    def create_address(self, public_key):
        unhashed_address = chr(1) + str(self.chain_id) + crypto.hash_chain(public_key.__bytes__())[0:20]
        address_hash = crypto.hash_chain(crypto.str2bytes(unhashed_address))[0:4]
        return base58.b58encode(crypto.str2bytes(unhashed_address + address_hash))

    def create_from_private_key(self, private_key):
        if not isinstance(private_key, SigningKey):
            public_key = VerifyKey(base58.b58decode(private_key)[-32:])
            private_key = SigningKey(seed=base58.b58decode(private_key)[:-32])
            address = self.create_address(public_key)
        else:
            public_key = private_key.verify_key
            address = self.create_address(public_key)
        return Account(address, public_key, private_key, self.key_type)

    def create_from_public_key(self, public_key):
        if isinstance(public_key, bytes):
            public_key = VerifyKey(public_key)
        elif isinstance(public_key, str):
            public_key = VerifyKey(base58.b58decode(public_key))

        if not isinstance(public_key, VerifyKey):
            raise Exception("Unrecognized Public Key format")

        address = self.create_address(public_key)

        return Account(address=address, public_key=public_key, key_type=self.key_type)

    def create_from_seed(self, seed, nonce=0):
        private_key, public_key, key_type = self.create_sign_keys(seed, nonce)
        address = self.create_address(public_key)
        return Account(address, public_key, private_key, key_type, seed, nonce)

    def create_with_values(self, address, public_key, private_key, key_type, seed=None):
        return Account(address, public_key, private_key, key_type, seed)




from LTO.Accounts.account_factory_ed25519 import AccountFactoryED25519
from LTO.public_node import PublicNode
from LTO.Transactions.anchor import Anchor
from LTO.Transactions.association import Association
from LTO.Transactions.revoke_association import RevokeAssociation



node = PublicNode('https://testnet.lto.network')
account = AccountFactoryED25519('T').create_from_seed('aaaaaaaaaaaaaaa')
account2 = AccountFactoryED25519('T').create_from_seed('bbbbbbbbbbbbbbb')
print(node.balance(account.address)/100000000)
print(node.balance(account2.address)/100000000)
anchor = '7ab62201df228f2c92ec74c29c61889b9658f4eef6a9a4a51bd25f23c9fcf376'
transaction = Anchor(anchor)
transaction.sign_with(account2)
transaction.sponsor_with(account)
tx = transaction.broadcast_to(node)







import base58
import struct
from LTO import crypto
from LTO.transaction import Transaction


class Transfer(Transaction):
    TYPE = 4
    DEFAULT_TX_FEE = 100000000
    DEFAULT_VERSION = 3

    def __init__(self, recipient, amount, attachment=''):
        super().__init__()
        self.recipient = recipient
        crypto.validate_address(recipient)
        self.amount = amount
        self.attachment = attachment
        self.version = self.DEFAULT_VERSION

        if self.amount <= 0:
            raise Exception('Amount should be positive')

        self.tx_fee = self.DEFAULT_TX_FEE

    def __to_binary_V2(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\2' +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.timestamp) +
                struct.pack(">Q", self.amount) +
                struct.pack(">Q", self.tx_fee) +
                base58.b58decode(self.recipient) +
                struct.pack(">H", len(self.attachment)) +
                crypto.str2bytes(self.attachment))

    def __to_binary_V3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                base58.b58decode(self.recipient) +
                struct.pack(">Q", self.amount) +
                struct.pack(">H", len(self.attachment)) +
                crypto.str2bytes(self.attachment))

    def to_binary(self):
        if self.version == 2:
            return self.__to_binary_V2()
        elif self.version == 3:
            return self.__to_binary_V3()
        else:
            raise Exception('Incorrect Version')

    def to_json(self):
        return ({
                    "type": self.TYPE,
                    "version": self.version,
                    "sender": self.sender,
                    "senderKeyType": self.sender_key_type,
                    "senderPublicKey": self.sender_public_key,
                    "fee": self.tx_fee,
                    "timestamp": self.timestamp,
                    "amount": self.amount,
                    "recipient": self.recipient,
                    "attachment": base58.b58encode(crypto.str2bytes(self.attachment)),
                    "proofs": self.proofs
                } | self._sponsor_json())

    @staticmethod
    def from_data(data):
        tx = Transfer(data['recipient'], data['amount'])
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.amount = data['amount']
        tx.recipient = data['recipient']
        tx.attachment = data['attachment'] if 'attachment' in data else ''
        tx.proofs = data['proofs'] if 'proofs' in data else []
        tx.height = data['height'] if 'height' in data else ''

        if 'sponsorPublicKey' in data:
            tx.sponsor = data['sponsor']
            tx.sponsor_public_key = data['sponsorPublicKey']

        return tx



import os
from LTO import crypto
from LTO.word_list import wordList
from abc import ABC, abstractmethod


class AccountFactory(ABC):

    def __init__(self, chain_id):
        self.chain_id = chain_id

    @abstractmethod
    def create_sign_keys(self, seed, nonce):
        pass

    @abstractmethod
    def create_address(self, public_key):
        pass

    def create(self):
        return self.create_from_seed(self.generate_seed())

    @abstractmethod
    def create_from_seed(self, seed, nonce=0):
        pass
        '''private_key, public_key, key_type = self.create_sign_keys(seed, nonce)
        address = self.create_address(public_key)
        return Account(address, public_key, private_key, key_type, seed, nonce)'''

    @abstractmethod
    def create_from_private_key(self, private_key):
        pass

    @abstractmethod
    def create_from_public_key(self, public_key):
        pass

    @abstractmethod
    def create_with_values(self, address, public_key, private_key, key_type, seed=''):
        pass

    def generate_seed(self):
        word_count = len(wordList)
        words = []
        for i in range(5):
            r = crypto.bytes2str(os.urandom(4))
            x = (ord(r[3])) + (ord(r[2]) << 8) + (ord(r[1]) << 16) + (ord(r[0]) << 24)
            w1 = x % word_count
            w2 = ((int(x / word_count) >> 0) + w1) % word_count
            w3 = ((int((int(x / word_count) >> 0) / word_count) >> 0) + w2) % word_count
            words.append(wordList[w1])
            words.append(wordList[w2])
            words.append(wordList[w3])
        return ' '.join(words)



import base64
from LTO.transaction import Transaction
import struct
from LTO import crypto
import base58
from LTO.public_node import PublicNode

class SetScript(Transaction):
    TYPE = 13
    DEFAULT_SCRIPT_FEE = 500000000
    DEFAULT_VERSION = 3

    def __init__(self, script):
        super().__init__()

        self.script = script
        if script:
            self.compiled_script = PublicNode('https://nodes.lto.network').compile(script)
        else:
            self.compiled_script = ""

        self.tx_fee = self.DEFAULT_SCRIPT_FEE
        self.version = self.DEFAULT_VERSION

    def __to_binary_V1(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\1' +
                crypto.str2bytes(crypto.get_network(self.sender)) +
                base58.b58decode(self.sender_public_key) +
                b'\1' +
                struct.pack(">H", len(self.compiled_script)) +
                self.compiled_script +
                struct.pack(">Q", self.tx_fee) +
                struct.pack(">Q", self.timestamp))

    def __to_binary_V3(self):
        return (self.TYPE.to_bytes(1, 'big') +
                b'\3' +
                crypto.str2bytes(self.chain_id) +
                struct.pack(">Q", self.timestamp) +
                crypto.key_type_id(self.sender_key_type) +
                base58.b58decode(self.sender_public_key) +
                struct.pack(">Q", self.tx_fee) +
                struct.pack(">H", len(base64.b64decode(self.compiled_script[7:]))) +
                base64.b64decode(self.compiled_script[7:])
                )

    def to_binary(self):
        if self.version == 1:
            return self.__to_binary_V1()
        elif self.version == 3:
            return self.__to_binary_V3()
        else:
            raise Exception('Incorrect Version')

    def to_json(self):
        return ({
            "type": self.TYPE,
            "version": self.version,
            "sender": self.sender,
            "senderKeyType": self.sender_key_type,
            "senderPublicKey": self.sender_public_key,
            "script": str(self.compiled_script),
            "timestamp": self.timestamp,
            "fee": self.tx_fee,
            "proofs": self.proofs
        } | self._sponsor_json())

    @staticmethod
    def from_data(data):
        tx = SetScript(data['script'])
        tx.id = data['id'] if 'id' in data else ''
        tx.type = data['type']
        tx.version = data['version']
        tx.sender = data['sender'] if 'sender' in data else ''
        tx.sender_key_type = data['senderKeyType'] if 'senderKeyType' in data else 'ed25519'
        tx.sender_public_key = data['senderPublicKey']
        tx.fee = data['fee']
        tx.timestamp = data['timestamp']
        tx.proofs = data['proofs']
        tx.script = data['script']
        tx.height = data['height'] if 'height' in data else ''
        return tx



from __future__ import absolute_import, division, print_function, unicode_literals

from LTO.account_factory import AccountFactory
from LTO.public_node import PublicNode

from LTO.Transactions.anchor import Anchor
from LTO.Transactions.lease import Lease
from LTO.Transactions.association import Association
from LTO.Transactions.cancel_lease import CancelLease
from LTO.Transactions.cancel_sponsorship import CancelSponsorship
from LTO.Transactions.mass_transfer import MassTransfer
from LTO.Transactions.revoke_association import RevokeAssociation
from LTO.Transactions.set_script import SetScript
from LTO.Transactions.sponsorship import Sponsorship
from LTO.Transactions.transfer import Transfer
from LTO.Accounts.account_factory_ecdsa import AccountFactoryECDSA
from LTO.Accounts.account_factory_ed25519 import AccountFactoryED25519


class PyCLTO:

    def __init__(self, chain_id='T'):

        if chain_id == 'T':
            self.NODE = PublicNode('https://testnet.lto.network')
        elif chain_id == 'L':
            self.NODE = PublicNode('https://nodes.lto.network')
        else:
            self.NODE = ''

        self.chain_id = chain_id

        self.account_factories = {
            'ed25519': AccountFactoryED25519(chain_id),
            'secp256r1': AccountFactoryECDSA(chain_id, curve='secp256r1'),
            'secp256k1': AccountFactoryECDSA(chain_id, curve='secp256k1')
        }

    def Account(self, public_key=None, private_key=None, key_type='ed25519', seed=None, nonce=0):
        factory = self.account_factories[key_type]
        if seed:
            account = factory.create_from_seed(seed, nonce)
        elif private_key:
            account = factory.create_from_private_key(private_key)
        elif public_key:
            account = factory.create_from_public_key(public_key)
        else:
            account = factory.create()
        return account

    def getchain_id(self):
        return self.chain_id

    def from_data(self, data):

        if data['type'] == 4:
            return Transfer(recipient=data['recipient'], amount=data['amount']).from_data(data)
        elif data['type'] == 8:
            return Lease(amount=1, recipient='').from_data(data)
        elif data['type'] == 11:
            return MassTransfer(transfers='').from_data(data)
        elif data['type'] == 15:
            return Anchor(anchor='').from_data(data)
        elif data['type'] == 16:
            return Association(recipient='', association_type='', anchor='').from_data(data)
        elif data['type'] == 17:
            return RevokeAssociation(recipient='', association_type='').from_data(data)
        elif data['type'] == 18:
            return Sponsorship(data['recipient']).from_data(data)
        elif data['type'] == 19:
            return CancelSponsorship(data['recipient']).from_data(data)
        elif data['type'] == 13:
            return SetScript(data['script']).from_data(data)
        elif data['type'] == 9:
            return CancelLease(leaseId='').from_data(data)
        else:
            raise Exception('Incorrect transaction Type')
